/*  You may distribute under the terms of either the GNU General Public License
 *  or the Artistic License (the same terms as Perl itself)
 *
 *  (C) Paul Evans, 2023 -- leonerd@leonerd.org.uk
 */
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "XSParseKeyword.h"

#include "perl-backcompat.c.inc"
#include "perl-additions.c.inc"

/* Wrappers between OP * and B::OP-blessed SVs */
#define newSVop(o)  S_newSVop(aTHX_ o)
static SV *S_newSVop(pTHX_ OP *o)
{
  SV *ret = newSV(0);

  const char *opclassname;
  switch(op_class(o)) {
    case OPclass_BASEOP:   opclassname = "B::OP";       break;
    case OPclass_UNOP:     opclassname = "B::UNOP";     break;
    case OPclass_BINOP:    opclassname = "B::BINOP";    break;
    case OPclass_LOGOP:    opclassname = "B::LOGOP";    break;
    case OPclass_LISTOP:   opclassname = "B::LISTOP";   break;
    case OPclass_PMOP:     opclassname = "B::PMOP";     break;
    case OPclass_SVOP:     opclassname = "B::SVOP";     break;
    case OPclass_PADOP:    opclassname = "B::PADOP";    break;
    case OPclass_PVOP:     opclassname = "B::PVOP";     break;
    case OPclass_LOOP:     opclassname = "B::LOOP";     break;
    case OPclass_COP:      opclassname = "B::COP";      break;
    case OPclass_METHOP:   opclassname = "B::METHOP";   break;
    case OPclass_UNOP_AUX: opclassname = "B::UNOP_AUX"; break;
    default:
      croak("TODO: handle opclass=%d\n", op_class(o));
  }

  sv_setiv(newSVrv(ret, opclassname), PTR2IV(o));
  return ret;
}

#define SvOPo(sv)  S_SvOPo(aTHX_ sv)
static OP *S_SvOPo(pTHX_ SV *sv)
{
  if(!SvROK(sv) || !sv_derived_from(sv, "B::OP"))
    croak("Expected a B::OP instance");

  return NUM2PTR(OP *, SvIV(SvRV(sv)));
}

struct XPKFPHookdata {
  /* Phase callbacks */
  CV *permitcv;
  CV *checkcv;
  CV *buildcv;

  SV *hookdata;
};

static bool cb_permit(pTHX_ void *hookdata)
{
  struct XPKFPHookdata *data = hookdata;

  dSP;

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  if(data->hookdata)
    XPUSHs(sv_mortalcopy(data->hookdata));
  else
    XPUSHs(&PL_sv_undef);
  PUTBACK;

  call_sv((SV *)data->permitcv, G_SCALAR);

  SPAGAIN;
  bool ret = SvTRUEx(POPs);

  FREETMPS;
  LEAVE;

  return ret;
}

static void cb_check(pTHX_ void *hookdata)
{
  struct XPKFPHookdata *data = hookdata;

  dSP;

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  if(data->hookdata)
    XPUSHs(sv_mortalcopy(data->hookdata));
  else
    XPUSHs(&PL_sv_undef);
  PUTBACK;

  call_sv((SV *)data->checkcv, G_VOID);

  FREETMPS;
  LEAVE;
}

static int cb_build(pTHX_ OP **out, XSParseKeywordPiece *args[], size_t nargs, void *hookdata)
{
  struct XPKFPHookdata *data = hookdata;

  if(nargs)
    croak("TODO: Wasn't expecting any args to cb_build but got %zd\n", nargs);

  dSP;
  SV *outsv = newSV(0);
  AV *argsav = newAV();

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  EXTEND(SP, 3);
  mPUSHs(newRV_noinc(outsv));
  mPUSHs(newRV_noinc((SV *)argsav));
  if(data->hookdata)
    PUSHs(sv_mortalcopy(data->hookdata));
  else
    PUSHs(&PL_sv_undef);
  PUTBACK;

  call_sv((SV *)data->buildcv, G_SCALAR);

  SPAGAIN;
  int ret = POPu;

  if(SvOK(outsv)) {
    *out = SvOPo(outsv);
  }

  FREETMPS;
  LEAVE;

  return ret;
}

#define ENTER_and_setup_pad(name)  S_ENTER_and_setup_pad(aTHX_ name)
static void S_ENTER_and_setup_pad(pTHX_ const char *name)
{
  if(!PL_compcv)
    croak("Cannot call %s while not compiling a subroutine", name);

  ENTER;

  PAD_SET_CUR(CvPADLIST(PL_compcv), 1);

  SAVESPTR(PL_comppad_name);
  PL_comppad_name = PadlistNAMES(CvPADLIST(PL_compcv));
}

static void S_setup_constants(pTHX)
{
  HV *stash;
  AV *export;

#define DO_CONSTANT(c) \
  newCONSTSUB(stash, #c, newSViv(c)); \
  av_push(export, newSVpv(#c, 0))

  stash = gv_stashpvs("XS::Parse::Keyword::FromPerl", TRUE);
  export = get_av("XS::Parse::Keyword::FromPerl::EXPORT_OK", TRUE);

  DO_CONSTANT(KEYWORD_PLUGIN_EXPR);
  DO_CONSTANT(KEYWORD_PLUGIN_STMT);

  DO_CONSTANT(OPf_WANT);
  DO_CONSTANT(OPf_WANT_VOID);
  DO_CONSTANT(OPf_WANT_SCALAR);
  DO_CONSTANT(OPf_WANT_LIST);
  DO_CONSTANT(OPf_KIDS);
  DO_CONSTANT(OPf_PARENS);
  DO_CONSTANT(OPf_REF);
  DO_CONSTANT(OPf_MOD);
  DO_CONSTANT(OPf_STACKED);
  DO_CONSTANT(OPf_SPECIAL);

  DO_CONSTANT(XPK_FLAG_EXPR);
  DO_CONSTANT(XPK_FLAG_STMT);
  DO_CONSTANT(XPK_FLAG_AUTOSEMI);
}

MODULE = XS::Parse::Keyword::FromPerl    PACKAGE = XS::Parse::Keyword::FromPerl

I32 opcode(const char *opname)
  CODE:
    for(RETVAL = 0; RETVAL < OP_max; RETVAL++)
      if(strEQ(opname, PL_op_name[RETVAL]))
        goto found;
    croak("Unrecognised opcode(\"%s\")", opname);
found:
    ;
  OUTPUT:
    RETVAL

SV *
newOP(I32 type, I32 flags)
  CODE:
    ENTER_and_setup_pad("newOP");
    RETVAL = newSVop(newOP(type, flags));
    LEAVE;
  OUTPUT:
    RETVAL

SV *
newUNOP(I32 type, I32 flags, SV *first)
  CODE:
    ENTER_and_setup_pad("newUNOP");
    RETVAL = newSVop(newUNOP(type, flags, SvOPo(first)));
    LEAVE;
  OUTPUT:
    RETVAL

SV *
newBINOP(I32 type, I32 flags, SV *first, SV *last)
  CODE:
    ENTER_and_setup_pad("newBINOP");
    RETVAL = newSVop(newBINOP(type, flags, SvOPo(first), SvOPo(last)));
    LEAVE;
  OUTPUT:
    RETVAL

SV *
newSVOP(I32 type, I32 flags, SV *sv)
  CODE:
    ENTER_and_setup_pad("newSVOP");
    RETVAL = newSVop(newSVOP(type, flags, newSVsv(sv)));
    LEAVE;
  OUTPUT:
    RETVAL

SV *
newGVOP(I32 type, I32 flags, SV *gv)
  CODE:
    if(!SvROK(gv) || SvTYPE(SvRV(gv)) != SVt_PVGV)
      croak("Expected a GLOB ref to newGVOP");
    ENTER_and_setup_pad("newGVOP");
    RETVAL = newSVop(newGVOP(type, flags, (GV *)SvRV(gv)));
    LEAVE;
  OUTPUT:
    RETVAL

void
register_xs_parse_keyword(const char *name, ...)
  CODE:
    dKWARG(1);

    struct XPKFPHookdata data = {0};
    U32 flags = 0;
    SV *permit_hintkeysv = NULL;

    static const char *args[] = {
      "flags",
      "pieces",
      "permit_hintkey",
      "permit",
      "check",
      /* TODO: parse? */
      "build",
      "hookdata",
    };
    while(KWARG_NEXT(args))
      switch(kwarg) {
        case 0: /* flags */
          flags = SvUV(kwval);
          break;

        case 1: /* pieces */
        {
          if(!SvROK(kwval) || SvTYPE(SvRV(kwval)) != SVt_PVAV)
            croak("Expected 'pieces' to be an array ref");
          AV *piecesav = AV_FROM_REF(kwval);
          if(av_count(piecesav))
            croak("TODO: Handle some pieces");

          break;
        }

        case 2: /* permit_hintkey */
          permit_hintkeysv = kwval;
          break;

        case 3: /* permit */
          if(!SvROK(kwval) || SvTYPE(SvRV(kwval)) != SVt_PVCV)
            croak("Expected 'permit' to be a CODE ref");
          data.permitcv = (CV *)SvREFCNT_inc((SV *)CV_FROM_REF(kwval));
          break;

        case 4: /* check */
          if(!SvROK(kwval) || SvTYPE(SvRV(kwval)) != SVt_PVCV)
            croak("Expected 'check' to be a CODE ref");
          data.checkcv = (CV *)SvREFCNT_inc((SV *)CV_FROM_REF(kwval));
          break;

        case 5: /* build */
          if(!SvROK(kwval) || SvTYPE(SvRV(kwval)) != SVt_PVCV)
            croak("Expected 'build' to be a CODE ref");
          data.buildcv = (CV *)SvREFCNT_inc((SV *)CV_FROM_REF(kwval));
          break;

        case 6: /* hookdata */
          data.hookdata = newSVsv(kwval);
          break;
      }

    if(!data.buildcv)
      croak("Require 'build' for register");
    if(!permit_hintkeysv && !data.permitcv)
      croak("Require at least one of 'permit_hintkey' or 'permit'");

    struct XSParseKeywordPieceType *pieces;
    Newx(pieces, 1, struct XSParseKeywordPieceType);
    *pieces = (struct XSParseKeywordPieceType){0};

    struct XSParseKeywordHooks *hooksptr;
    Newx(hooksptr, 1, struct XSParseKeywordHooks);

    *hooksptr = (struct XSParseKeywordHooks){
      .flags  = flags,
      .pieces = pieces,
    };
    if(permit_hintkeysv)
      hooksptr->permit_hintkey = savepv(SvPV_nolen(permit_hintkeysv));
    if(data.permitcv)
      hooksptr->permit = &cb_permit;
    if(data.checkcv)
      hooksptr->check = &cb_check;
    if(data.buildcv)
      hooksptr->build = &cb_build;

    struct XPKFPHookdata *dataptr;
    Newx(dataptr, 1, struct XPKFPHookdata);
    *dataptr = data;

    register_xs_parse_keyword(savepv(name), hooksptr, dataptr);

BOOT:
  boot_xs_parse_keyword(0);

  S_setup_constants(aTHX);
